@@ -2,6 +2,7 @@ package Mojo::IOLoop::ReadWriteFork;
 use Mojo::Base 'Mojo::EventEmitter';
 
 use Errno qw(EAGAIN ECONNRESET EINTR EPIPE EWOULDBLOCK EIO);
+use IO::Handle;
 use IO::Pty;
 use Mojo::Asset::Memory;
 use Mojo::IOLoop;
@@ -13,8 +14,10 @@ use Scalar::Util qw(blessed);
 
 use constant CHUNK_SIZE => $ENV{MOJO_CHUNK_SIZE} || 131072;
 use constant DEBUG      => $ENV{MOJO_READWRITEFORK_DEBUG} && 1;
+use constant READ       => 0;
+use constant WRITE      => 1;
 
-our $VERSION = '1.02';
+our $VERSION = '2.00';
 
 our @SAFE_SIG
   = grep { !m!^(NUM\d+|__[A-Z0-9]+__|ALL|CATCHALL|DEFER|HOLD|IGNORE|MAX|PAUSE|RTMAX|RTMIN|SEGV|SETS)$! } keys %SIG;
@@ -29,7 +32,7 @@ sub close {
   my $self = shift;
   my $what = $_[0] eq 'stdout' ? 'stdout_read' : 'stdin_write';    # stdout_read is EXPERIMENTAL
   my $fh   = delete $self->{$what} or return $self;
-  CORE::close($fh) or $self->emit(error => $!);
+  $fh->close or $self->emit(error => $!);
   $self;
 }
 
@@ -74,28 +77,32 @@ sub start {
 
 sub _start {
   my ($self, $args) = @_;
-  my ($stdin_read, $stdin_write, $stdout_read, $stdout_write);
+  my (@stdin, @stdout, @stderr);
 
   local $!;
   if ($args->{conduit} eq 'pipe') {
-    pipe $stdout_read, $stdout_write or return $self->emit(error => "pipe: $!");
-    pipe $stdin_read,  $stdin_write  or return $self->emit(error => "pipe: $!");
-    select +(select($stdout_write), $| = 1)[0];
-    select +(select($stdin_write),  $| = 1)[0];
+    @stdin  = $self->_pipe;
+    @stdout = $self->_pipe;
   }
   elsif ($args->{conduit} eq 'pty') {
-    $stdin_write = $stdout_read = IO::Pty->new;
+    $stdin[WRITE] = $stdout[READ] = IO::Pty->new;
   }
   else {
     warn "[RWF] Invalid conduit ($args->{conduit})\n" if DEBUG;
     return $self->emit(error => "Invalid conduit ($args->{conduit})");
   }
 
+  if ($args->{stderr}) {
+    @stderr = $self->_pipe;
+  }
+
   my $prepare_event = {
-    stdin_read   => $stdin_read,
-    stdin_write  => $stdin_write,
-    stdout_read  => $stdout_read,
-    stdout_write => $stdout_write,
+    stderr_read  => $stderr[READ],
+    stderr_write => $stderr[WRITE],
+    stdin_read   => $stdin[READ],
+    stdin_write  => $stdin[WRITE],
+    stdout_read  => $stdout[READ],
+    stdout_write => $stdout[WRITE],
   };
 
   $self->emit(before_fork => $prepare_event);    # LEGACY
@@ -103,29 +110,31 @@ sub _start {
 
   return $self->emit(error => "Couldn't fork ($!)") unless defined($self->{pid} = fork);
   return $self->{pid}
-    ? $self->_start_parent($args, $stdin_read, $stdin_write, $stdout_read, $stdout_write)
-    : $self->_start_child($args, $stdin_read, $stdin_write, $stdout_read, $stdout_write);
+    ? $self->_start_parent($args, \@stdin, \@stdout, \@stderr)
+    : $self->_start_child($args, \@stdin, \@stdout, \@stderr);
 }
 
 sub _start_child {
-  my ($self, $args, $stdin_read, $stdin_write, $stdout_read, $stdout_write) = @_;
+  my ($self, $args, $stdin, $stdout, $stderr) = @_;
 
-  if (blessed $stdin_write and $stdin_write->isa('IO::Pty')) {
-    $stdin_write->make_slave_controlling_terminal;
-    $stdin_read = $stdout_write = $stdin_write->slave;
-    $stdin_read->set_raw                                         if $args->{raw};
-    $stdin_read->clone_winsize_from($args->{clone_winsize_from}) if $args->{clone_winsize_from};
+  if (blessed $stdin->[WRITE] and $stdin->[WRITE]->isa('IO::Pty')) {
+    $stdin->[WRITE]->make_slave_controlling_terminal;
+    $stdin->[READ] = $stdout->[WRITE] = $stdin->[WRITE]->slave;
+    $stdin->[READ]->set_raw                                         if $args->{raw};
+    $stdin->[READ]->clone_winsize_from($args->{clone_winsize_from}) if $args->{clone_winsize_from};
   }
 
-  CORE::close($stdin_write);
-  CORE::close($stdout_read);
-  open STDIN,  '<&' . fileno $stdin_read   or exit $!;
-  open STDOUT, '>&' . fileno $stdout_write or exit $!;
-  open STDERR, '>&' . fileno $stdout_write or exit $!;
-  select STDERR;
-  $| = 1;
-  select STDOUT;
-  $| = 1;
+  my $stdout_no = ($args->{stdout} // 1) && fileno($stdout->[WRITE]);
+  my $stderr_no = ($args->{stderr} // 1) && fileno($stderr->[WRITE] || $stdout->[WRITE]);
+  open STDIN,  '<&' . fileno($stdin->[READ]) or exit $!;
+  open STDOUT, '>&' . $stdout_no             or exit $! if $stdout_no;
+  open STDERR, '>&' . $stderr_no             or exit $! if $stderr_no;
+  $stdout_no ? STDOUT->autoflush(1) : STDOUT->close;
+  $stderr_no ? STDERR->autoflush(1) : STDERR->close;
+
+  $stdin->[WRITE]->close;
+  $stdout->[READ]->close;
+  $stderr->[READ]->close if $stderr->[READ];
 
   %ENV = %{$args->{env}};
 
@@ -146,25 +155,17 @@ sub _start_child {
 }
 
 sub _start_parent {
-  my ($self, $args, $stdin_read, $stdin_write, $stdout_read, $stdout_write) = @_;
+  my ($self, $args, $stdin, $stdout, $stderr) = @_;
 
   $self->_d("Forked $args->{program} @{$args->{program_args} || []}") if DEBUG;
-  @$self{qw(stdin_write stdout_read)} = ($stdin_write, $stdout_read);
-  @$self{qw(wait_eof wait_sigchld)}   = (1, 1);
-  $stdout_read->close_slave if blessed $stdout_read and $stdout_read->isa('IO::Pty');
+  @$self{qw(stdin_write stdout_read stderr_read)} = ($stdin->[WRITE], $stdout->[READ], $stderr->[READ]);
+  @$self{qw(wait_eof wait_sigchld)}               = (1, 1);
 
-  my $stream = Mojo::IOLoop::Stream->new($stdout_read)->timeout(0);
-  $stream->on(error => sub { $! != EIO && $self->emit(error => "Read error: $_[1]") });
-  $stream->on(close => sub { $self->_maybe_terminate('wait_eof') });
-  $stream->on(
-    read => sub {
-      $self->_d(sprintf ">>> RWF\n%s", term_escape $_[1]) if DEBUG;
-      $self->emit(read => $_[1]);
-    }
-  );
+  $stdout->[READ]->close_slave if blessed $stdout->[READ] and $stdout->[READ]->isa('IO::Pty');
+  $self->_stream($args, stderr => $stderr->[READ]) if $stderr->[READ];
+  $self->_stream($args, stdout => $stdout->[READ]) if !$stderr->[READ] or $args->{stdout};
 
   $SIGCHLD->waitpid($self->{pid} => sub { $self->_sigchld(@_) });
-  $self->{stream_id} = $self->ioloop->stream($stream);
   $self->emit('fork');    # LEGACY
   $self->emit('spawn');
   $self->_write;
@@ -210,6 +211,27 @@ sub _maybe_terminate {
   $self->emit(error => $_) for @errors;
 }
 
+sub _on_read_cb {
+  my ($self, $args, $name) = @_;
+  my @emit = $args->{stderr} ? ($name) : qw(read);
+  return sub { $self->emit(@emit => $_[1]) }
+    unless DEBUG;
+
+  my $NAME = uc $name;
+  return sub {
+    $self->_d(sprintf ">>> RWF:$NAME\n%s", term_escape $_[1]) if DEBUG;
+    $self->emit(@emit => $_[1]);
+  };
+}
+
+sub _pipe {
+  my $self = shift;
+  my @rw;
+  pipe $rw[READ], $rw[WRITE] or return $self->emit(error => "pipe: $!");
+  $rw[WRITE]->autoflush(1);
+  return @rw;
+}
+
 sub _sigchld {
   my ($self, $status, $pid) = @_;
   my ($exit_value, $signal) = ($status >> 8, $status & 127);
@@ -218,6 +240,15 @@ sub _sigchld {
   $self->_maybe_terminate('wait_sigchld');
 }
 
+sub _stream {
+  my ($self, $args, $name, $handle) = @_;
+  my $stream = Mojo::IOLoop::Stream->new($handle)->timeout(0);
+  $stream->on(error => sub { $! != EIO && $self->emit(error => "Read error: $_[1]") });
+  $stream->on(close => sub { $self->_maybe_terminate('wait_eof') });
+  $stream->on(read  => $self->_on_read_cb($args, $name));
+  $self->ioloop->stream($stream);
+}
+
 sub _write {
   my $self = shift;
 
@@ -249,7 +280,7 @@ Mojo::IOLoop::ReadWriteFork - Fork a process and read/write from it
 
 =head1 VERSION
 
-1.02
+2.00
 
 =head1 SYNOPSIS
 
@@ -294,6 +325,20 @@ more than welcome.
 
 =head1 EVENTS
 
+=head2 stderr
+
+  $fork->on(stderr => sub { my ($fork, $buf) = @_; });
+
+Emitted when the child has written a chunk of data to STDERR and L</conduit>
+has the "stderr" key set to a true value.
+
+=head2 stdout
+
+  $fork->on(stdout => sub { my ($fork, $buf) = @_; });
+
+Emitted when the child has written a chunk of data to STDOUT and L</conduit>
+has the "stdout" key set to a true value.
+
 =head2 asset
 
   $fork->on(asset => sub { my ($fork, $asset) = @_; });
@@ -330,7 +375,8 @@ Emitted when the child process exit.
 
   $fork->on(read => sub { my ($fork, $buf) = @_; });
 
-Emitted when the child has written a chunk of data to STDOUT or STDERR.
+Emitted when the child has written a chunk of data to STDOUT or STDERR, and
+neither "stderr" nor "stdout" is set in the L</conduit>.
 
 =head2 spawn
 
@@ -364,13 +410,17 @@ See also L</pid> for example usage of this event.
 Emitted right before the child process is forked. Example C<$pipes>
 
   $pipes = {
+    # if "stderr" is set in conduit()
+    stdin_write => $stderr_fh_w,
+    stdout_read => $stderr_fh_r,
+
     # for both conduit "pipe" and "pty"
-    stdin_write => $pipe_fh_1_or_pty_object,
-    stdout_read => $pipe_fh_2_or_pty_object,
+    stdin_write => $pipe_fh_r_or_pty_object,
+    stdout_read => $pipe_fh_w_or_pty_object,
 
     # only for conduit "pipe"
-    stdin_read => $pipe_fh_3,
-    stdout_write => $pipe_fh_4,
+    stdin_read   => $pipe_fh_r,
+    stdout_write => $pipe_fh_w,
   }
 
 =head1 ATTRIBUTES
@@ -378,11 +428,31 @@ Emitted right before the child process is forked. Example C<$pipes>
 =head2 conduit
 
   $hash = $fork->conduit;
-  $fork = $fork->conduit({type => "pipe"});
+  $fork = $fork->conduit(\%options);
 
-Used to set the conduit and conduit options. Example:
+Used to set the conduit options. Possible values are:
 
-  $fork->conduit({raw => 1, type => "pty"});
+=over 2
+
+=item * stderr
+
+This will make L<Mojo::IOLoop::ReadWriteFork> emit "stderr" events, instead of
+"read" events. Setting this to "0" will close STDERR in the child.
+
+=item * stdout
+
+This will make L<Mojo::IOLoop::ReadWriteFork> emit "stdout" events, instead of
+"read" events. Setting this to "0" will close STDOUT in the child.
+
+=item * raw
+
+Calls L<IO::Pty/set_raw> if "typ" is "pty".
+
+=item * type
+
+"type" can be either "pipe" or "pty". Default value is "pipe".
+
+=back
 
 =head2 ioloop
 
